unit untAuthClient;

(*******************************************************************************

Simle socket client with SSPI authentication
Author: Sergey N. Naberegnyh

Version 1.2.0.0
Created: November, 16, 2008
Updated: -  December, 03, 2008

*******************************************************************************)

interface

uses
  Windows, SysUtils, WinSock, Variants,
  untSharedTypes, JwaSspi, JwaWinError,
  StrUtils, HPSockApi, SockSSPIMics;

procedure ClientMain;

const
  IDs_DOMAIN_NAME      = 1000;
  IDs_VERIFICATION_OK   = 1001;
  IDs_VERIFICATION_FAIL = 1002;
  IDs_INPUT_CMD         = 1003;
  IDs_INPUT_PARAMS      = 1004;
  IDs_ACCEPT_RESPONSE   = 1005;
  IDs_RESPONSE          = 1006;
  IDs_DONE_1            = 1007;
  IDs_DONE_2            = 1008;
  IDs_DONE_3            = 1009;
  IDs_USAGE             = 1010;
  IDs_PARAM_1           = 1011;
  IDs_PARAM_2           = 1012;
  IDs_PARAM_3           = 1013;
  IDs_EXAMPLE_1         = 1014;
  IDs_EXAMPLE_2         = 1015;
  IDs_EXAMPLE_3         = 1016;
  IDs_AUTHOR            = 1017;
  IDs_INPUT_USER        = 1018;
  IDs_INPUT_PASSWORD    = 1019;
  IDs_ACCESS_DENIED     = 1020;
  IDs_CONNECTION_CLOSED = 1021;

implementation

var
  InBuf, OutBuf: array[0..$7FFF] of Byte;
  UserName: AnsiString;
  ServerLogonName: PSecChar = nil;
  CredInteractive: boolean;
  GuestEnabled: boolean;

function CredUIPrompt(Domain: PWideChar;
  var UserBuf; UserBufSize: DWORD; var PswdBuf; PswdBufSize: DWORD): Cardinal;
var
  s: AnsiString;
  Mode, OldMode: Cardinal;
  StdIn: THandle;
  n: integer;
  c: Char;
  P: PWideChar;
  L: integer;
begin
  WriteOem(LoadStrRes(IDs_INPUT_USER));
  ReadLn(s);
  if s = '' then Result:= ERROR_CANCELLED
  else if Length(s) >= integer(UserBufSize div SizeOf(WideChar)) then
    Result:= ERROR_INSUFFICIENT_BUFFER
  else Result:= NO_ERROR;

  if Result <> NO_ERROR then Exit;

  OemToCharW(@s[1], @UserBuf);

  WriteOem(LoadStrRes(IDs_INPUT_PASSWORD));
  StdIn:= GetStdHandle(STD_INPUT_HANDLE);
  GetConsoleMode(StdIn, OldMode);
  Mode:= OldMode and not (ENABLE_LINE_INPUT or ENABLE_ECHO_INPUT);
  SetConsoleMode(StdIn, Mode);

  n:= 0;
  P:= @PswdBuf;
  L:= integer(PswdBufSize div SizeOf(WideChar));
  repeat
    Read(C);
    FlushConsoleInputBuffer(StdIn);
    if C = #13 then Break;
    OemToCharBuffW(@C, P, 1);
    C:=#0;
    Inc(n);
    Inc(P);
  until n >= L;
  if n >= L then
  begin
    FillChar(PswdBuf, PswdBufSize, 0);
    Result:= ERROR_INSUFFICIENT_BUFFER;
  end else
    P^:= #0;
  WriteLn;  
  SetConsoleMode(StdIn, OldMode);
end;

function GetCredAttributes(var Domain; DomainBufSize: DWORD;
  var UserBuf; UserBufSize: DWORD; var PswdBuf; PswdBufSize: DWORD): boolean;
var
  hLib: HModule;
  s: AnsiString;
  SysCredUIPrompt: TCredUICmdLinePromptForCredentialsW;
  R: Cardinal;
begin
  WideChar(UserBuf):= #0;
  WideChar(PswdBuf):= #0;
  WriteOem(LoadStrRes(IDs_DOMAIN_NAME));
  ReadLn(s);
  if (s <> '') and (Length(s) < integer(DomainBufSize div SizeOf(WideChar)))
  then  OemToCharW(@s[1], @Domain)
  else WideChar(Domain):= #0;

  hLib:= LoadLibrary('credui.dll');
  if 0 <> hLib then
    @SysCredUIPrompt:= GetProcAddress(hLib, 'CredUICmdLinePromptForCredentialsW')
  else
    @SysCredUIPrompt:= nil;

  if nil <> @SysCredUIPrompt then
    R:= SysCredUIPrompt(
              @Domain, nil, 0, @UserBuf, UserBufSize div SizeOf(WideChar),
              @PswdBuf, PswdBufSize div SizeOf(WideChar), nil,
              CREDUI_FLAGS_DO_NOT_PERSIST or
              CREDUI_FLAGS_EXCLUDE_CERTIFICATES or
              CREDUI_FLAGS_GENERIC_CREDENTIALS)
  else
    R:= CredUIPrompt(@Domain, UserBuf, UserBufSize, PswdBuf, PswdBufSize);
    
  if 0 <> hLib then FreeLibrary(hLib);
  Result:= R = NO_ERROR;
  if not Result then WritelnOem(['Cred. prompt failed: ', SysErrorMessage(R)]);
end;
     
function SignBuffer(const SecHandle: TSecHandle; PData: PByte;
  DataLen: integer; var SignSize: Longint; var SignBuf): boolean;
var
  Desc: SecBufferDesc;
  SecBuf: array[0..2] of SecBuffer;
  SS: SECURITY_STATUS;
begin
  with Desc do
  begin
    ulVersion:= SECBUFFER_VERSION;
    cBuffers:= 3;
    pBuffers:= @SecBuf;
  end;

  with SecBuf[0] do
  begin
    cbBuffer:= SignSize;
    BufferType:= SECBUFFER_TOKEN;
    pvBuffer:= @SignBuf;
  end;

  with SecBuf[1] do
  begin
    cbBuffer:= DataLen;
    BufferType:= SECBUFFER_DATA;
    pvBuffer:= PData;
  end;

  with SecBuf[2] do
  begin
    cbBuffer:= 0;
    BufferType:= SECBUFFER_PADDING;
    pvBuffer:= nil;
  end;

  SS:= MakeSignature(@SecHandle, 0, @Desc, 0);
  Result:= Succeeded(SS);
  if Result then SignSize:= SecBuf[0].cbBuffer;
end;

function CheckSign(const SecHandle: TSecHandle; PData: PByte;
  DataLen: integer; out PResponse: PAnsiChar; out ResponseLen: integer): boolean;
var
  Desc: SecBufferDesc;
  SecBuf: array[0..1] of SecBuffer;
  QOP: Cardinal;
  SS: SECURITY_STATUS;
begin
  QOP:= 0;
  with Desc do
  begin
    ulVersion:= SECBUFFER_VERSION;
    cBuffers:= 2;
    pBuffers:= @SecBuf;
  end;

  with SecBuf[0] do
  begin
    cbBuffer:= PDWORD(PData)^;
    BufferType:= SECBUFFER_TOKEN;
    pvBuffer:= Pointer(Cardinal(PData) + SizeOf(DWORD));
  end;

  PResponse:= Pointer(Cardinal(PData) + SecBuf[0].cbBuffer + SizeOf(DWORD));
  ResponseLen:= DataLen - integer(SecBuf[0].cbBuffer) - SizeOf(DWORD);
  with SecBuf[1] do
  begin
    cbBuffer:= ResponseLen;
    BufferType:= SECBUFFER_DATA;
    pvBuffer:= PResponse;
  end;

  SS:= VerifySignature(@SecHandle, @desc, 0, QOP);
  Result:= Succeeded(SS);
  if Result then
    WriteLnOem(LoadStrRes(IDs_VERIFICATION_OK))
  else
    WritelnOem([LoadStrRes(IDs_VERIFICATION_FAIL), SysErrorMessage(SS)]);
end;

procedure DoWorkLoop(ClientSocket: TSocket; const SecHandle: TSecHandle);
var
  sCmd, sParam: AnsiString;
  PCmd: PCmdHeader;
  P: PChar;
  Datalen: Longint;
  Sizes: SecPkgContext_Sizes;
  ss: SECURITY_STATUS;
  PSign: PByte;
  SignSize: Longint;
  WsaBuf: array[0..3] of TWsaBuf;
  BytesTransfered: Cardinal;
  FirstLoop: boolean;
begin
  FirstLoop:= true;
  
  ss:= QueryContextAttributes(@SecHandle, SECPKG_ATTR_SIZES, @Sizes);
  if Failed(SS) then
  begin
    WriteLnOem(['QueryContextAttributes failed: ', SysErrorMessage(SS)]);
    Exit;
  end;
  GetMem(PSign, Sizes.cbMaxSignature);
  try
    repeat

      if SizeOf(Datalen) <> recv(ClientSocket, DataLen, SizeOf(Datalen), 0) then
      begin
        WriteLnOem(['Receiving failed: ', SysErrorMessage(WSAGetLastError)]);
        Break;
      end;

      if Datalen = 0 then
      begin
        if FirstLoop then WritelnOem(LoadStrRes(IDs_ACCESS_DENIED))
        else WritelnOem(LoadStrRes(IDs_CONNECTION_CLOSED));
        Break;
      end;

      if DataLen > SizeOf(InBuf) then
      begin
        WriteLn('Inbound message too long: ', DataLen);
        Break;
      end;

      if datalen <> recv(ClientSocket, InBuf, Datalen, 0) then
      begin
        WriteLnOem(['Receiving failed: ', SysErrorMessage(WSAGetLastError)]);
        Break;
      end;

      if not FirstLoop then WriteLnOem(LoadStrRes(IDs_ACCEPT_RESPONSE));
      if not CheckSign(SecHandle, @InBuf, Datalen, P, Datalen) then Break;
      WriteLn;
      SetString(sCmd, P, Datalen);
      if not FirstLoop then WriteLnOem(LoadStrRes(IDs_RESPONSE));
      WriteLnOem(sCmd);
      WriteLn;

      FirstLoop:= false;

      WriteOem(LoadStrRes(IDs_INPUT_CMD));
      Readln(sCmd);
      if sCmd = '' then Break;
      sCmd:= UpperCase(sCmd);
      WriteOem(LoadStrRes(IDs_INPUT_PARAMS));
      ReadLn(sParam);
      
      PCmd:= @OutBuf;
      P:= Pointer(PCmd);
      Inc(P, SizeOf(PCmd^));
      PCmd.CmdOffset:= SizeOf(PCmd^);
      StrPCopy(P, sCmd);
      PCmd.CmdLen:= Length(sCmd);
      Inc(P, PCmd.CmdLen);
      PCmd.ParamOffset:= PCmd.CmdOffset + PCmd.CmdLen;
      PCmd.ParamLen:= Length(sParam);
      StrPCopy(P, sParam);

      SignSize:= Sizes.cbMaxSignature;
      DataLen:= SizeOf(PCmd^) + PCmd.CmdLen + PCmd.ParamLen;
      if not SignBuffer(SecHandle, Pointer(PCmd), Datalen, SignSize, PSign^) then
      begin
        Writeln('Signing failed');
        Break;
      end;

      WsaBuf[1].cLength:= SizeOf(SignSize);
      WsaBuf[1].pBuffer:= @SignSize;
      WsaBuf[2].cLength:= SignSize;
      Wsabuf[2].pBuffer:= PSign;
      WsaBuf[3].cLength:= Datalen;
      WsaBuf[3].pBuffer:= Pointer(PCmd);
      Datalen:= WsaBuf[1].cLength + WsaBuf[2].cLength + WsaBuf[3].cLength;
      WsaBuf[0].cLength:= SizeOf(DataLen);
      WsaBuf[0].pBuffer:= @DataLen;
      if 0 <> WsaSend(ClientSocket, WsaBuf[0], 4,
                       BytesTransfered, 0, nil, nil) then
      begin
        WriteLnOem(['Sending failed: ', SysErrorMessage(WSAGetLastError)]);
        break;
      end;

    until false;
  finally
    Freemem(PSign);
  end;
end;

function CreataCredentialHandle(out CredHandle: TCredHandle): boolean;
var
  ss: SECURITY_STATUS;
  Expiry: TimeStamp;
  AuthIdent: TSecWinNTAuthIdentityW;
  PAuthident: PSecWinNTAuthIdentityW;
  wDomain: array[0..CRED_MAX_DOMAIN_TARGET_NAME_LENGTH] of WideChar;
  BufName: array[0..CRED_MAX_USERNAME_LENGTH] of WideChar;
  BufPswd: array[0..127] of WideChar;
begin
  Result:= false;

  if CredInteractive then
  begin

    if not GetCredAttributes(
                  wDomain, SizeOf(wDomain),
                  BufName, SizeOf(BufName),
                  BufPswd, SizeOf(BufPswd)) then Exit;

    with AuthIdent do
    begin
      User:= @BufName;
      UserLength:= WStrLen(AuthIdent.User);
      Password:= @BufPswd;
      PasswordLength:= WStrLen(AuthIdent.Password);
      Domain:= @wDomain;
      DomainLength:= WStrLen(AuthIdent.Domain);
      Flags:= SEC_WINNT_AUTH_IDENTITY_UNICODE;
    end;
    PAuthIdent:= @AuthIdent;
  end else
    PAuthident:= nil;

  SS:= AcquireCredentialsHandle(nil, @sPackageName[1],
         SECPKG_CRED_OUTBOUND, nil, PAuthIdent, nil, nil, @CredHandle, Expiry);
  FillChar(BufPswd, SizeOf(BufPswd), 0);

  Result:= Succeeded(SS);
  if Result then WriteLn else
    WriteLnOem(['AcquireCredentialsHandle failed: ', SysErrorMessage(SS)]);
end;

function DoClientAuth(ClientSocket: TSocket;
  out CredHandle: TCredHandle; out SecHandle: TSecHandle): boolean;
var
  ss: SECURITY_STATUS;
  Expiry: TimeStamp;
  InDesc, OutDesc: SecBufferDesc;
  InSecBuf, OutSecBuf: SecBuffer;
  PInDesc: PSecBufferDesc;
  Attrib: ULONG;
  WsaBuf: array[0..1] of TWsaBuf;
  BytesTransfered: Cardinal;
  NegInfo: SecPkgContext_NegotiationInfo;
  PSH: PCtxtHandle;
  Flags: Cardinal;
begin
  Result:= false;
  BytesTransfered:= 0;
  FillChar(SecHandle, SizeOf(SecHandle), 0);
  if GuestEnabled then Flags:= 0
  else Flags:= ISC_REQ_INTEGRITY;

  if CreataCredentialHandle(CredHandle) then
  try
    repeat
      with OutDesc do
      begin
        ulVersion:= SECBUFFER_VERSION;
        cBuffers:= 1;
        pBuffers:= @OutSecBuf;
      end;

      with OutSecBuf do
      begin
        cbBuffer:= SizeOf(OutBuf);
        BufferType:= SECBUFFER_TOKEN;
        pvBuffer:= @OutBuf;
      end;

      if BytesTransfered <> 0 then
      begin
        with InDesc do
        begin
          ulVersion:= SECBUFFER_VERSION;
          cBuffers:= 1;
          pBuffers:= @InSecBuf;
        end;

        with InSecBuf do
        begin
          cbBuffer:= BytesTransfered;
          BufferType:= SECBUFFER_TOKEN;
          pvBuffer:= @InBuf;
        end;

        PSH:= @SecHandle;
        PInDesc:= @InDesc;
      end else
      begin  // first iteration
        PSH:= nil;
        PInDesc:= nil;
      end;

      ss:= InitializeSecurityContext(@CredHandle, PSH,
             ServerLogonName, Flags, 0,
             SECURITY_NATIVE_DREP, PInDesc,
             0, @SecHandle, @OutDesc, Attrib, @Expiry);

      if Failed(SS) then
      begin
        WriteLnOem(['InitializeSecurityContext failed: ', SysErrorMessage(SS)]);
        Exit;
      end;

      if (SEC_I_COMPLETE_NEEDED = ss) or (SEC_I_COMPLETE_AND_CONTINUE = ss) then
      begin
        ss:= CompleteAuthToken(@SecHandle, @OutDesc);
        if Failed(SS) then
        begin
          WriteLnOem(['CompleteAuthToken failed: ', SysErrorMessage(SS)]);
          Exit;
        end;
      end;

      if OutSecBuf.cbBuffer <> 0 then
      begin
        WsaBuf[0].cLength:= SizeOf(DWORD);
        WsaBuf[0].pBuffer:= @OutSecBuf.cbBuffer;
        WsaBuf[1].cLength:= OutSecBuf.cbBuffer;
        WsaBuf[1].pBuffer:= OutSecBuf.pvBuffer;
        if 0 <> WSASend(ClientSocket, WsaBuf[0], 2,
                          BytesTransfered, 0, nil, nil) then
        begin
          WriteLnOem(['Sending failed: ', SysErrorMessage(WSAGetLastError)]);
          Exit;
        end;
      end;

      if SEC_E_OK = SS then
      begin
        Result:= true;
        SS:= QueryContextAttributes(@SecHandle, SECPKG_ATTR_NEGOTIATION_INFO, @NegInfo);
        if Succeeded(SS) then
        try
          WriteLnOem(['Negotiation package: ', StrPas(@NegInfo.PackageInfo.Name^)]);
        finally
          FreeContextBuffer(NegInfo.PackageInfo);
        end;
        Break;
      end;

      if SizeOf(BytesTransfered) <> Recv(ClientSocket, BytesTransfered,
                                              SizeOf(BytesTransfered), 0) then
      begin
        WriteLnOem(['Receiving failed: ', SysErrorMessage(WSAGetLastError)]);
        Exit;
      end;

      if BytesTransfered > SizeOf(InBuf) then
      begin
        WriteLn('Inbound message too long: ', BytesTransfered);
        Exit;
      end;

      if integer(BytesTransfered) <>
                  Recv(ClientSocket, InBuf, BytesTransfered, 0) then
      begin
        WriteLnOem(['Receiving failed: ', SysErrorMessage(WSAGetLastError)]);
        Exit;
      end;

    until false;
    
  finally
    if not Result then
    begin
      if (SecHandle.dwLower or SecHandle.dwUpper) <> 0 then
        DeleteSecurityContext(@SecHandle);
      FreeCredentialHandle(@CredHandle);
    end;
  end;
end;

function ParseCmdLine(out pAddr: PAddrInfo): boolean;
var
  s: string;
  Count, n: integer;
  sServer, sPort: string;
begin
  CredInteractive:= false;
  GuestEnabled:= false;
  Count:= ParamCount;
  Result:= Count >= 2;
  if not Result then Exit;

  sServer:= ParamStr(Count - 1);
  sPort:= ParamStr(Count);

  Result:= false;

  for n:= 1 to Count - 2 do
  begin
    s:= ParamStr(n);
    if (Length(s) < 2) or (not (s[1] in ['/', '-'])) then Exit;

    case s[2] of
      'i':
      begin
        if Length(s) > 2 then Exit;
        if UserName <> '' then Exit;
        CredInteractive:= true;
      end;

      'g':
      begin
        if Length(s) > 2 then Exit;
        GuestEnabled:= true;
      end;

      'u':
      begin
        if CredInteractive or (Length(s) < 3) then Exit;
        UserName:= Copy(s, 3, MaxInt);
      end;

      else Exit;
    end;
  end;

  if UserName <> '' then ServerLogonName:= @UserName[1];
  try
    pAddr:= ResolveAddressAndPortV6(false, sServer, sPort, AF_INET, SOCK_STREAM);
  except
    on E: ESockAddrError do
    begin
      WriteLnOem(['Address resolution failed: ', E.Message]);
      Exit;
    end;
    else raise;
  end;
  Result:= true;
end;

{$HINTS OFF}
procedure Final;
var
  C: Char;
  OldMode, Mode: DWORD;
  StdIn: THandle;
begin
  repeat
    if Redirected then Break;
    StdIn:= GetStdHandle(STD_INPUT_HANDLE);
    if INVALID_HANDLE_VALUE = StdIn then Break;
    if GetConsoleMode(StdIn, OldMode) then
    begin
      Mode:= OldMode and not (ENABLE_LINE_INPUT or ENABLE_ECHO_INPUT);
      if SetConsoleMode(StdIn, Mode) then
      begin
        WriteOem(LoadStrRes(IDs_DONE_1));
        Read(C);
        SetConsoleMode(StdIn, OldMode);
      end else
      begin
        WriteOem(LoadStrRes(IDs_DONE_2));
        ReadLn;
      end;
      Exit;
    end;
  until true;
  
  WriteLnOem(LoadStrRes(IDs_DONE_3));
end;
{$HINTS ON}

procedure Usage();
begin
  WriteLnOem(LoadStrRes(IDs_USAGE));
  WriteLn;
  WriteLnOem(LoadStrRes(IDs_PARAM_1));
  WriteLnOem(LoadStrRes(IDs_PARAM_2));
  WriteLnOem(LoadStrRes(IDs_PARAM_3));
  WriteLn;
  WriteLnOem(LoadStrRes(IDs_EXAMPLE_1));
  WriteLnOem(LoadStrRes(IDs_EXAMPLE_2));
  WriteLnOem(LoadStrRes(IDs_EXAMPLE_3));
  WriteLn;
end;

procedure ClientMain;
var
  R: integer;
  ClientSocket: TSocket;
  pAddr: PAddrInfo;
  WsaData: TWSAData;
  CredHandle: TCredHandle;
  SecHandle: TSecHandle;
begin
  try
    Write(#$D#$A);
    Writeln('Authenticated Socket Client.');
    WriteLnOem(LoadStrRes(IDs_AUTHOR));
    WriteLn;

    R:= WSAStartup(MakeWord(2, 0), WsaData);
    if 0 <> R then
    begin
      WriteLn('Could not initialize winsock subsistem');
      WriteLnOem(['Error: ', SysErrorMessage(R)]);
      Exit;
    end;

    try

      if not ParseCmdLine(pAddr) then
      begin
        Usage();
        Exit;
      end;

      try

        ClientSocket:= socket(pAddr.ai_family, pAddr.ai_socktype, 0);
        if INVALID_SOCKET =  ClientSocket then
        begin
          WriteLnOem(['Creation socket failed: ', SysErrorMessage(WSAGetLastError)]);
          Exit;
        end;

        try
          if 0 <> connect(ClientSocket, pAddr.ai_addr^, pAddr.ai_addrlen) then
          begin
            WriteLnOem(['Connection failed: ', SysErrorMessage(WSAGetLastError)]);
            Exit;
          end;

          if DoClientAuth(ClientSocket, CredHandle, SecHandle) then
          try
            DoWorkLoop(ClientSocket, SecHandle);
          finally
            DeleteSecurityContext(@SecHandle);
            FreeCredentialHandle(@CredHandle);
          end;
        finally
          closesocket(ClientSocket);
        end;

      finally
        freeaddrinfo(pAddr);
      end;

    finally
      WSACleanup;
    end;
  finally
    Final;
  end;
end;

end.
