unit untAuthSrvrMain;

(*******************************************************************************

Simle socket server with SSPI authentication
Author: Sergey N. Naberegnyh

Version 1.2.0.0
Created: November, 16, 2008
Updated: -  December, 03, 2008

*******************************************************************************)

interface

uses
  Windows, Messages, SysUtils, Variants,
  Classes, Graphics, Controls, Forms,
  Dialogs, HPScktSrvr, StdCtrls, JwaSSPI, WinSock;

type
  ESSPIException = class(Exception);

  TServerMainForm = class(TForm)
    Label1: TLabel;
    Label2: TLabel;
    edPort: TEdit;
    edInterface: TEdit;
    btnOpen: TButton;
    btnClose: TButton;
    HPServerSocket1: THPServerSocket;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure HPServerSocket1ClientBeforeAccept(
      AClient: TCustomHPServerClient; var ConnBufSize: Integer);
    procedure HPServerSocket1ClientConnect(AClient: TCustomHPServerClient;
      pConnectionData: PByte; ConnectionDataLen: Integer);
    procedure HPServerSocket1ClientDisconnect(
      AClient: TCustomHPServerClient);
    procedure HPServerSocket1ReadComplete(AClient: TCustomHPServerClient;
      BytesTransfered: Cardinal; CompletionKey, Error: Integer);
    procedure HPServerSocket1WriteComplete(AClient: TCustomHPServerClient;
      BytesTransfered: Cardinal; CompletionKey, Error: Integer);
    procedure btnOpenClick(Sender: TObject);
    procedure btnCloseClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure HPServerSocket1CreateAcceptor(Sender: TCustomHPServerSocket;
      Socket: Integer);
    procedure HPServerSocket1CreateListener(Sender: TCustomHPServerSocket;
      Socket: Integer);
  private
    { Private declarations }
    procedure EnableControls(Enable: boolean);
  public
    { Public declarations }
  end;

var
  ServerMainForm: TServerMainForm;

const
  cMaxMsgSize: Cardinal = 0;
  hCredHandle: TCredHandle = ();

implementation

uses
  untSharedTypes, untSspiSrvClient;

{$R *.dfm}

procedure TServerMainForm.EnableControls(Enable: boolean);
begin
  edPort.Enabled:= Enable;
  edInterface.Enabled:= Enable;
  btnOpen.Enabled:= Enable;
  btnClose.Enabled:= not Enable;
end;

procedure TServerMainForm.FormCreate(Sender: TObject);
var
  pPkgInfo: PSecPkgInfo;
  Expiry: TimeStamp;
  SS: SECURITY_STATUS;
begin
  SS:= QuerySecurityPackageInfo(@sPackageName[1], pPkgInfo);
  if Failed(SS) then
  begin
    Application.MessageBox(PAnsiChar(SysErrorMessage(ss)), nil, MB_ICONERROR);
    Application.Terminate;
  end;
  cMaxMsgSize:= pPkgInfo.cbMaxToken;
  FreeContextBuffer(pPkgInfo);

  SS:= AcquireCredentialsHandle(nil, @sPackageName[1],
         SECPKG_CRED_INBOUND, nil, nil, nil, nil, @hCredHandle, Expiry);
  if Failed(SS) then
  begin
    Application.MessageBox(PAnsiChar(SysErrorMessage(ss)), nil, MB_ICONERROR);
    Application.Terminate;
  end;
end;

procedure TServerMainForm.FormDestroy(Sender: TObject);
begin
  if (hCredHandle.dwLower or hCredHandle.dwUpper) <> 0 then
    FreeCredentialHandle(@hCredHandle);
end;

procedure TServerMainForm.HPServerSocket1ClientBeforeAccept(
  AClient: TCustomHPServerClient; var ConnBufSize: Integer);
begin
  ConnBufSize := 2 * cMaxMsgSize;
end;

procedure TServerMainForm.HPServerSocket1ClientConnect(
  AClient: TCustomHPServerClient; pConnectionData: PByte;
  ConnectionDataLen: Integer);
begin
  TSspiSrvClient(AClient).Initialize;
  TSspiSrvClient(AClient).DoReadComplete(DWORD(ConnectionDataLen), 0);
end;

procedure TServerMainForm.HPServerSocket1ClientDisconnect(
  AClient: TCustomHPServerClient);
begin
  TSspiSrvClient(AClient).Finalize;
end;

procedure TServerMainForm.HPServerSocket1ReadComplete(
  AClient: TCustomHPServerClient; BytesTransfered: Cardinal; CompletionKey,
  Error: Integer);
begin
  if Error = 0 then
    TSspiSrvClient(AClient).DoReadComplete(BytesTransfered, CompletionKey);
end;

procedure TServerMainForm.HPServerSocket1WriteComplete(
  AClient: TCustomHPServerClient; BytesTransfered: Cardinal; CompletionKey,
  Error: Integer);
begin
  if Error = 0 then
    TSspiSrvClient(AClient).DoWriteComplete(BytesTransfered, CompletionKey);
end;

procedure TServerMainForm.btnOpenClick(Sender: TObject);
begin
  HPServerSocket1.Address:= edInterface.Text;
  HPServerSocket1.ServiceOrPort:= edPort.Text;
  HPServerSocket1.ClientClass:= TSspiSrvClient;
  HPServerSocket1.Open;
  EnableControls(false);
end;

procedure TServerMainForm.btnCloseClick(Sender: TObject);
begin
//  HPServerSocket1.EnumerateConnections(EnumClients);
  HPServerSocket1.Close(10000);
  EnableControls(true);
end;

procedure TServerMainForm.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
  btnCloseClick(nil);
end;

procedure TServerMainForm.HPServerSocket1CreateAcceptor(
  Sender: TCustomHPServerSocket; Socket: Integer);
var
  B: bool;
begin
  B:= true;
  setsockopt(Socket, SOL_SOCKET, SO_REUSEADDR, @B, sizeof(B));
end;

procedure TServerMainForm.HPServerSocket1CreateListener(
  Sender: TCustomHPServerSocket; Socket: Integer);
var
  B: bool;
begin
  B:= true;
  setsockopt(Socket, SOL_SOCKET, SO_REUSEADDR, @B, sizeof(B));
end;

end.
