unit untHttpMain;

(*******************************************************************************

Simle HTTP(S) server
Author: Sergey N. Naberegnyh

Version 2.1.2.0
Created: August, 05, 2008
Updated: December, 03, 2008

*******************************************************************************)

{$IFDEF LOGGING_ON}
{$IFOPT D-}
{$DEFINE INTERNAL_LOG}
{$ENDIF}
{$ENDIF}

{$WARN UNIT_PLATFORM OFF}

interface

uses
  Windows, Messages, SysUtils, Variants,
  Classes, Graphics, Controls, Forms,
  Dialogs, SockSSPIMics, HPSockApi, HPScktSrvr,
  WinSock, StdCtrls, Buttons, ExtCtrls, FileCtrl,
  CommCtrl;

type
  TBalloonIcon = (TIP_NOICON = 0, TIP_ICONINFO, TIP_ICONWARNING, TIP_ICONERROR);

  TGarbageThread = class(TThread)
  private
    FServer: THPServerSocket;
    FTerminateEvent: THandle;
    function EnumClients(ACient: TCustomHPServerClient): boolean;
  protected
    procedure Execute; override;
  public
    constructor Create(AServer: THPServerSocket);
    destructor Destroy; override;
    procedure Terminate;
  end;

  TForm1 = class(TForm)
    HPServerSocket1: THPServerSocket;
    btnOpen: TButton;
    btnClose: TButton;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    edRootDir: TEdit;
    edInterface: TEdit;
    edPort: TEdit;
    btnSelectDir: TBitBtn;
    Bevel1: TBevel;
    chbUseSSL: TCheckBox;
    gbxSSLOptions: TGroupBox;
    edCertName: TEdit;
    Label4: TLabel;
    GroupBox1: TGroupBox;
    chbSSL2: TCheckBox;
    chbSSL3: TCheckBox;
    chbTLS1: TCheckBox;
    chbStoreLocation: TCheckBox;
    Bevel2: TBevel;
    chbPCT1: TCheckBox;
    chbMutualAuth: TCheckBox;
    procedure HPServerSocket1ClientBeforeAccept(
      AClient: TCustomHPServerClient; var ConnBufSize: Integer);
    procedure HPServerSocket1ClientConnect(AClient: TCustomHPServerClient;
      pConnectionData: PByte; ConnectionDataLen: Integer);
    procedure btnOpenClick(Sender: TObject);
    procedure btnCloseClick(Sender: TObject);
    procedure HPServerSocket1ClientDisconnect(
      AClient: TCustomHPServerClient);
    procedure HPServerSocket1ReadComplete(AClient: TCustomHPServerClient;
      BytesTransfered: Cardinal; CompletionKey, Error: Integer);
    procedure HPServerSocket1WriteComplete(AClient: TCustomHPServerClient;
      BytesTransfered: Cardinal; CompletionKey, Error: Integer);
    procedure btnSelectDirClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure HPServerSocket1ThreadStart(Sender: TObject);
    procedure HPServerSocket1ThreadEnd(Sender: TObject);
    procedure chbUseSSLClick(Sender: TObject);
    procedure edRootDirEnter(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormClick(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
  private
    { Private declarations }
    FGarbageThread: TGarbageThread;
    FHintWnd: HWND;
{$IFDEF INTERNAL_LOG}
    FLogs: TThreadList;
    procedure SaveInternalLog();
{$ENDIF}
{$IFDEF LOGGING_ON}
    procedure ServerLogMessage(const Params: array of PAnsiChar;
      EventType, Category: Word; ID: Cardinal);
{$ENDIF}
    procedure EnableControls(Enable: boolean);
    procedure ShowBallonNotification(Control: TControl;
      Icon: TBalloonIcon; const Title: string; const Message: WideString;
      BalloonColor: integer; MaxWidth: integer = 0);
    procedure AppDeactivate(Sender: TObject);
    procedure HideBalloon;
    procedure WmWindowPosChanging(var Message); message WM_WINDOWPOSCHANGING;
  public
    { Public declarations }
  end;

var
  Form1: TForm1;
  RootDir: string = '.\WWW';

{$IFDEF INTERNAL_LOG}
threadvar
  ThreadLog: TStringList;
{$ENDIF}

const
  TTS_BALLOON = $40;
  TTM_ADDTOOL = $0400 + 50;
  TTM_SETTITLE = WM_USER + 32;

  sSend_OK = 'HTTP/1.0 200 OK'#$0D#$0A'Server: HttpDemo 2.0'#$0D#$0A +
    'Content-length: %d'#$0D#$0A#$0D#$0A;
  sSend_OK_KeepAlive = 'HTTP/1.0 200 OK'#$0D#$0A'Server: HttpDemo 2.0'#$0D#$0A +
    'Connection: Keep-Alive'#$0D#$0A +
    'Content-length: %d'#$0D#$0A#$0D#$0A;

  MAX_LINE_LEN = 1000;
  WSATYPE_NOT_FOUND = 10109;

function NotifyBalloonHint(HintWnd: HWND; Icon: TBalloonIcon;
  const Title: AnsiString; const Message: WideString;
  const ScreenPos: TPoint; Center: boolean; MaxWidth: integer = 0;
  BackColor: COLORREF = $00CDF7F8; CaptColor: COLORREF = 0): HWND;

procedure HideBalloonHint(hHint: HWND);

implementation

uses
  untClient, untSslClient;

{$R *.dfm}
{$R WindowsXP.res}

function NotifyBalloonHint(HintWnd: HWND; Icon: TBalloonIcon;
  const Title: AnsiString; const Message: WideString;
  const ScreenPos: TPoint; Center: boolean; MaxWidth: integer = 0;
  BackColor: COLORREF = $00CDF7F8; CaptColor: COLORREF = 0): HWND;
var
  Info: TOOLINFOW;
begin
  if HintWnd <> 0 then Result:= HintWnd else
    Result:= CreateWindow(TOOLTIPS_CLASS, nil,
                          WS_POPUP or TTS_NOPREFIX or TTS_BALLOON,
                          0, 0, 0, 0, 0, 0, HInstance, nil);

  if 0 = Result then Exit;

  FillChar(Info, SizeOf(Info), 0);
  Info.cbSize := SizeOf(Info);
  Info.uId:= Result;
  Info.uFlags := TTF_TRANSPARENT or TTF_SUBCLASS or TTF_TRACK;
  if Center then Info.uFlags:= Info.uFlags or TTF_CENTERTIP;
  Info.lpszText := PWideChar(Message);

  if HintWnd <> 0 then
    SendMessage(Result, TTM_DELTOOL, 0, Integer(@Info));

  SetWindowPos(Result, HWND_TOPMOST, 0, 0, 0, 0,
                 SWP_NOACTIVATE or SWP_NOMOVE or SWP_NOSIZE);
  if 0 <> MaxWidth then
    SendMessage(Result, TTM_SETMAXTIPWIDTH, 0, MaxWidth);
  SendMessage(Result, TTM_SETTIPBKCOLOR, BackColor, 0);
  SendMessage(Result, TTM_SETTIPTEXTCOLOR, CaptColor, 0);
  SendMessage(Result, TTM_ADDTOOL, 0, Integer(@Info));
  SendMessage(Result, TTM_SETTITLE, Ord(Icon), Integer(PAnsiChar(Title)));
  SendMessage(Result, TTM_TRACKPOSITION, 0, MakeLong(ScreenPos.X, ScreenPos.Y));
  SendMessage(Result, TTM_TRACKACTIVATE, integer(true), integer(@Info));
end;

procedure HideBalloonHint(hHint: HWND);
var
  Info: TOOLINFOW;
begin
  if hHint = 0 then Exit;
  FillChar(Info, SizeOf(Info), 0);
  Info.cbSize := SizeOf(Info);
  Info.uId:= hHint;
  SendMessage(hHint, TTM_TRACKACTIVATE, integer(false), integer(@Info));
end;

{ TForm1 }

procedure TForm1.HPServerSocket1ClientBeforeAccept(
  AClient: TCustomHPServerClient; var ConnBufSize: Integer);
begin
  ConnBufSize:= $10000 - 2 * Addr_Buf_Len;
end;

{$IFDEF LOGGING_ON}
var
  LogLineID: integer;
{$ENDIF}

procedure TForm1.btnOpenClick(Sender: TObject);
var
  Protocols: Cardinal;
  w: integer;
begin
  HideBalloon;
  ClientIDs:= 0;
  RootDir:= edRootDir.Text;
  if DirectoryExists(RootDir) then
  begin
    HPServerSocket1.Address:= edInterface.Text;
    HPServerSocket1.ServiceOrPort:= edPort.Text;
  end else
  begin
    ShowBallonNotification(edRootDir, TIP_ICONERROR, 'Error!',
     'This directory not exists!'#13#10'Directory name must be valid!', $00EAE2FC);
    Exit;
  end;
{$IFDEF LOGGING_ON}
  HPServerSocket1.OnLogMessage:= ServerLogMessage;
  LogLineID:= 0;
{$ENDIF}
  if chbUseSSL.Checked then
  begin
    Protocols:= 0;
    if chbSSL2.Checked then Protocols:= Protocols or SP_PROT_SSL2;
    if chbSSL3.Checked then Protocols:= Protocols or SP_PROT_SSL3;
    if chbTLS1.Checked then Protocols:= Protocols or SP_PROT_TLS1;
    if chbPCT1.Checked then Protocols:= Protocols or SP_PROT_PCT1;
    try
      SslSecurity:= TSSL_Security.Create(edCertName.Text,
                  chbStoreLocation.Checked, chbMutualAuth.Checked, Protocols);
    except
      on E: ESslCertStoreException do
      begin
        w:= edCertName.Width + edCertName.Width div 2;
        ShowBallonNotification(chbStoreLocation, TIP_ICONERROR,
                             'Error!', E.Message, $00E2D7FB, w);
        Exit;
      end;

      on E: ESslCertException do
      begin
        w:= edCertName.Width + edCertName.Width div 2;
        ShowBallonNotification(edCertName, TIP_ICONERROR,
                             'Error!', E.Message, $00E2D7FB, w);
        Exit;
      end;

      on E: ESslException do
      begin
        ShowBallonNotification(chbUseSSL, TIP_ICONERROR,
                             'Error!', E.Message, $00E2D7FB);
        Exit;
      end;

      else raise;
    end;

    HPServerSocket1.ClientClass:= THttpSrvClientSSL;
  end else
  begin
    HPServerSocket1.ClientClass:= THttpSrvClient;
//    DigestSecurity:= TDigestSecurity.Create(atDigest);
    DigestSecurity:= TDigestSecurity.Create;
  end;

  FGarbageThread:= TGarbageThread.Create(HPServerSocket1);

  try
    HPServerSocket1.Open;
  except
    on E: Exception do
    begin
      if Assigned(FGarbageThread) then
      begin
        FGarbageThread.Terminate;
        FreeAndNil(FGarbageThread);
      end;
      FreeAndNil(SslSecurity);

      if E is ESockAddrError then
      begin
        if (WSATYPE_NOT_FOUND = ESockAddrError(E).ErrCode) or
                        (WSANO_DATA = ESockAddrError(E).ErrCode) then
          ShowBallonNotification(edPort, TIP_ICONERROR, 'Error',
                                        'Unknown service name!', $00E2D7FB)
        else
          ShowBallonNotification(edInterface, TIP_ICONERROR, 'Error',
                                                      E.Message, $00E2D7FB);
        Exit;                                              
      end else
        raise;
    end;
  end;

  EnableControls(false);
end;

procedure TForm1.btnCloseClick(Sender: TObject);
begin
  if Assigned(FGarbageThread) then
  begin
    FGarbageThread.Terminate;
    FreeAndNil(FGarbageThread);
  end;
  HPServerSocket1.Close(INFINITE);  //(3000);
  FreeAndNil(SslSecurity);
  FreeAndNil(DigestSecurity);
  EnableControls(true);
{$IFDEF INTERNAL_LOG}
  SaveInternalLog;
{$ENDIF}  
end;

procedure TForm1.HPServerSocket1ClientConnect(
  AClient: TCustomHPServerClient; pConnectionData: PByte;
  ConnectionDataLen: Integer);
var
  ID: integer;
begin
  ID:= THttpSrvClient(AClient).ClientID;
  HPServerSocket1.LogMessage(['--->Connect'], 0, 0, Cardinal(ID));
  with THttpSrvClient(AClient) do
  begin
    Initialize;
    ProcessBuffer(PAnsiChar(pConnectionData), ConnectionDataLen);
  end;
end;

procedure TForm1.HPServerSocket1ClientDisconnect(
  AClient: TCustomHPServerClient);
var
  ID: integer;
begin
  ID:= THttpSrvClient(AClient).ClientID;
  HPServerSocket1.LogMessage(['--->Disonnect'], 0, 0, Cardinal(ID));
  THttpSrvClient(AClient).Finalize;;
end;

procedure TForm1.HPServerSocket1ReadComplete(
  AClient: TCustomHPServerClient; BytesTransfered: Cardinal; CompletionKey,
  Error: Integer);
begin
  HPServerSocket1.LogMessage(['--->ReadComplete', PAnsiChar(IntToStr(BytesTransfered))], 0, 0, Cardinal(THttpSrvClient(AClient).ClientID));
  if Error <> 0 then AClient.Disconnect else
    with THttpSrvClient(AClient) do
      ProcessBuffer(PAnsiChar(Buffer), integer(BytesTransfered));
end;

procedure TForm1.HPServerSocket1WriteComplete(
  AClient: TCustomHPServerClient; BytesTransfered: Cardinal; CompletionKey,
  Error: Integer);
begin
  THttpSrvClient(AClient).DoWriteComplete(BytesTransfered, CompletionKey, Error);
end;

procedure TForm1.btnSelectDirClick(Sender: TObject);
var
  s: string;
begin
  HideBalloon;
  s:= edRootDir.Text;
  if SelectDirectory('Root directory', '', s) then
    edRootDir.Text:= s;
end;

procedure TForm1.EnableControls(Enable: boolean);
begin
  btnOpen.Enabled:= Enable;
  btnClose.Enabled:= not Enable;
  edRootDir.Enabled:= Enable;
  btnSelectDir.Enabled:= Enable;
  edInterface.Enabled:= Enable;
  edPort.Enabled:= Enable;
  chbUseSSL.Enabled:= Enable;
  gbxSSLOptions.Enabled:= Enable;
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  chbUseSSLClick(nil);
  edRootDir.Text:= RootDir;
  Application.OnDeactivate:= AppDeactivate;
{$IFDEF LOGGING_ON}
  Caption:= 'Logging ';
{$ENDIF}
{$IFDEF INTERNAL_LOG}
  FLogs:= TThreadList.Create;
  Caption:= Caption + 'Internal';
{$ENDIF}
end;

{$IFDEF LOGGING_ON}
procedure TForm1.ServerLogMessage(const Params: array of PAnsiChar;
  EventType, Category: Word; ID: Cardinal);
var
  s, st: string;
  n: integer;
begin
  n:= InterLockedIncrement(LogLineID);
  st:= TimeToStr(Time);
  case length(Params) of
    1: s:= Format('[%.4d]=%s= $%p: %s', [n, st, Pointer(ID), Params[0]]);
    2: s:= Format('[%.4d]=%s= $%p: %s -> %s', [n, st, Pointer(ID), Params[0], Params[1]]);
    else s:= Format('[%.4d]=%s= $%p', [n, st, Pointer(ID)]);
  end;
{$IFDEF INTERNAL_LOG}
  ThreadLog.Add(s);
{$ELSE}
  OutputDebugString(PAnsiChar(s));
{$ENDIF}
end;
{$ENDIF}

procedure TForm1.FormDestroy(Sender: TObject);
{$IFDEF INTERNAL_LOG}
var
  n: integer;
{$ENDIF}
begin
  if FHintWnd <> 0 then DestroyWindow(FHintWnd);
  btnCloseClick(nil);
{$IFDEF INTERNAL_LOG}
  with FLogs.LockList do
  try
    for n:= 0 to Pred(Count) do
      TObject(Items[n]).Free;
  finally
    FLogs.UnlockList;
  end;
  FLogs.Free;
{$ENDIF}
end;

procedure TForm1.HPServerSocket1ThreadStart(Sender: TObject);
begin
{$IFDEF INTERNAL_LOG}
  ThreadLog:= TStringList.Create;
{$ENDIF}
end;

procedure TForm1.HPServerSocket1ThreadEnd(Sender: TObject);
begin
{$IFDEF INTERNAL_LOG}
  FLogs.Add(ThreadLog);
  ThreadLog:= nil;
{$ENDIF}
end;

{$IFDEF INTERNAL_LOG}
procedure TForm1.SaveInternalLog;
var
  n: integer;
  Stm: TFileStream;
  s: string;
begin
  with FLogs.LockList do
  try
    if Count = 0 then Exit;
  finally
    FLogs.UnlockList;
  end;

  Stm:= TFileStream.Create('InternalLog.txt', fmCreate);
  try
    with FLogs.LockList do
    try
      for n:= 0 to Pred(Count) do
      begin
        s:= Format('#####################################= %d =#####################################'#13#10, [n]);
        Stm.Write(s[1], Length(s));
        TStrings(Items[n]).SaveToStream(Stm);
        TObject(Items[n]).Free;
        Items[n]:= nil;
      end;
      Clear;
    finally
      FLogs.UnlockList;
    end;
  finally
    Stm.Free;
  end;
end;
{$ENDIF}

procedure TForm1.chbUseSSLClick(Sender: TObject);
begin
  if chbUseSSL.Checked then
  begin
    ClientWidth:= gbxSSLOptions.Left + gbxSSLOptions.Width + 6;
    gbxSSLOptions.Visible:= true;
    Bevel2.Visible:= true;
  end else
  begin
    Bevel2.Visible:= false;
    gbxSSLOptions.Visible:= false;
    ClientWidth:= Bevel2.Left + Bevel2.Width - 2; 
  end;
end;

procedure TForm1.edRootDirEnter(Sender: TObject);
begin
  HideBalloon;
end;

procedure TForm1.ShowBallonNotification(Control: TControl;
  Icon: TBalloonIcon; const Title: string; const Message: WideString;
  BalloonColor: integer; MaxWidth: integer);
var
  P: TPoint;
begin
  if Control = nil then Exit;
  P:= Control.ClientToScreen(Point(4, edRootDir.Height - 8));
  FHintWnd:= NotifyBalloonHint(FHintWnd, Icon, Title,
                        Message, P, false, MaxWidth, BalloonColor);
  KeyPreview:= true;
  Beep;
end;

procedure TForm1.WmWindowPosChanging(var Message);
begin
  HideBalloon;
  inherited;
end;

procedure TForm1.HideBalloon;
begin
  KeyPreview:= false;
  HideBalloonHint(FHintWnd);
end;

procedure TForm1.AppDeactivate(Sender: TObject);
begin
  if 0 <> FHintWnd then
  begin
    ShowWindow(FHintWnd, SW_HIDE);
    HideBalloon;
  end;
end;

{ TGarbageThread }

constructor TGarbageThread.Create(AServer: THPServerSocket);
begin
  FTerminateEvent:= CreateEvent(nil, false, false, nil);
  FServer:= AServer;
  inherited Create(false);
end;

destructor TGarbageThread.Destroy;
begin
  CloseHandle(FTerminateEvent);
  inherited;
end;

function TGarbageThread.EnumClients(
  ACient: TCustomHPServerClient): boolean;
begin
  Result:= not Terminated;
  if not Result then Exit;
  if ACient.ConnectionTime > 10 then
  begin
    if ACient is THttpSrvClientSSL then
      with THttpSrvClientSSL(ACient) do
        if ClientState > SSL_NEGOTIATEFAIL then SSLDisconnect;

    ACient.Disconnect;
  end;
end;

procedure TGarbageThread.Execute;
begin
{$IFDEF INTERNAL_LOG}
  ThreadLog:= TStringList.Create;
  try
{$ENDIF}
  repeat
    if WaitForSingleObject(FTerminateEvent, 60000) = WAIT_TIMEOUT then
      FServer.EnumerateConnections(EnumClients) else Break;

    if DigestSecurity = nil then Continue;

    while DigestSecurity.CheckSessionsLiveTime(GetTickCount, 2 * 60000) do
    begin
      if Terminated then Break;
      Sleep(0);
    end;
  until Terminated;
{$IFDEF INTERNAL_LOG}
  finally
    Form1.FLogs.Add(ThreadLog);
    ThreadLog:= nil;
  end;
{$ENDIF}
end;

procedure TGarbageThread.Terminate;
begin
  inherited;
  SetEvent(FTerminateEvent);
end;

procedure TForm1.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  HideBalloon;
end;

procedure TForm1.FormClick(Sender: TObject);
begin
  HideBalloon;
end;

procedure TForm1.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
  if FHintWnd <> 0 then
  begin
    DestroyWindow(FHintWnd);
    FHintWnd:= 0;
  end;
end;

end.

