unit SockSSPIMics;

interface

uses
  Windows, Messages, SysUtils, WinSock,
  JwaSspi, JwaWinCrypt, JwaWinError;

const
  WS2_32_LIB = 'WS2_32.dll';

  UNISP_NAME_A = 'Microsoft Unified Security Protocol Provider';

  SCHANNEL_CRED_VERSION           = $00000004;

  SCH_CRED_NO_DEFAULT_CREDS       = $00000010;
  SCH_CRED_MANUAL_CRED_VALIDATION = $00000008;
  
  CREDUI_FLAGS_DO_NOT_PERSIST              = $00002;
  CREDUI_FLAGS_EXCLUDE_CERTIFICATES        = $00008;
  CREDUI_FLAGS_GENERIC_CREDENTIALS         = $40000;

  CRED_MAX_DOMAIN_TARGET_NAME_LENGTH       = (256+1+80);
  CRED_MAX_USERNAME_LENGTH                 = (256+1+256);
  SECPKG_ATTR_CONNECTION_INFO     = $0000005A;
  SECPKG_ATTR_ISSUER_LIST_EX      = $00000059;
  SECPKG_ATTR_REMOTE_CERT_CONTEXT = $00000053;
  
  SP_PROT_PCT1_SERVER           = $00000001;
  SP_PROT_PCT1_CLIENT           = $00000002;
  SP_PROT_PCT1                  = (SP_PROT_PCT1_SERVER or SP_PROT_PCT1_CLIENT);

  SP_PROT_SSL2_SERVER           = $00000004;
  SP_PROT_SSL2_CLIENT           = $00000008;
  SP_PROT_SSL2                  = (SP_PROT_SSL2_SERVER or SP_PROT_SSL2_CLIENT);

  SP_PROT_SSL3_SERVER           = $00000010;
  SP_PROT_SSL3_CLIENT           = $00000020;
  SP_PROT_SSL3                  = (SP_PROT_SSL3_SERVER or SP_PROT_SSL3_CLIENT);

  SP_PROT_TLS1_SERVER           = $00000040;
  SP_PROT_TLS1_CLIENT           = $00000080;
  SP_PROT_TLS1                  = (SP_PROT_TLS1_SERVER or SP_PROT_TLS1_CLIENT);

  SP_PROT_SSL3TLS1_CLIENTS      = (SP_PROT_TLS1_CLIENT or SP_PROT_SSL3_CLIENT);
  SP_PROT_SSL3TLS1_SERVERS      = (SP_PROT_TLS1_SERVER or SP_PROT_SSL3_SERVER);

  SCHANNEL_SHUTDOWN               = 1;
  IO_BUFFER_SIZE                  =  $10000;

  DLL_NAME                        = 'Secur32.dll';
  NT4_DLL_NAME                    = 'Security.dll';

type
  _SCHANNEL_CRED = packed record
     dwVersion: DWORD;
     cCreds: DWORD;
     paCred: PCCERT_CONTEXT;
     hRootStore: HCERTSTORE;
     cMappers: DWORD;
     aphMappers: PPointer; //struct _HMAPPER**;
     cSupportedAlgs: DWORD;
     palgSupportedAlgs: ^ALG_ID;
     grbitEnabledProtocols: DWORD;
     dwMinimumCipherStrength: DWORD;
     dwMaximumCipherStrength: DWORD;
     dwSessionLifespan: DWORD;
     dwFlags: DWORD;
     reserved: DWORD;
  end;
  SCHANNEL_CRED = _SCHANNEL_CRED;
  PSCHANNEL_CRED = ^SCHANNEL_CRED;

  _SecPkgContext_ConnectionInfo = packed record
    dwProtocol: DWORD;
    aiCipher: ALG_ID;
    dwCipherStrength: DWORD;
    aiHash: ALG_ID;
    dwHashStrength: DWORD;
    aiExch: ALG_ID;
    dwExchStrength: DWORD;
  end;
  SecPkgContext_ConnectionInfo = _SecPkgContext_ConnectionInfo;
  PSecPkgContext_ConnectionInfo = ^_SecPkgContext_ConnectionInfo;

  _SecPkgContext_IssuerListInfoEx = packed record
    aIssuers: PCERT_NAME_BLOB;
    cIssuers: DWORD;
  end;
  SecPkgContext_IssuerListInfoEx = _SecPkgContext_IssuerListInfoEx;
  PSecPkgContext_IssuerListInfoEx = ^_SecPkgContext_IssuerListInfoEx;

  TCredUICmdLinePromptForCredentialsW = function
     (pszTargetName: LPCWSTR; pContext: PCtxtHandle;
     dwAuthError: DWORD; UserName: PWideChar; ulUserBufferSize: ULONG;
     pszPassword: PWideChar; ulPasswordBufferSize: ULONG;
     pfSave: PBOOL; dwFlags: DWORD): DWORD; stdcall;

procedure WriteOem(const Value: Variant); overload;
procedure WriteOem(const Params: array of Variant); overload;
procedure WriteLnOem(const Value: Variant); overload;
procedure WriteLnOem(const Params: array of Variant); overload;

function WStrLen(PW: PWideChar): integer;
function LoadStrRes(ResID: integer): string;
function Redirected(): boolean;

implementation

procedure WriteOem(const Value: Variant); overload;
var
  s: string;
begin
  s:= Value;
  if s <> '' then
  begin
    CharToOemBuff(@s[1], @s[1], Length(s));
    Write(s);
  end;
end;

procedure WriteOem(const Params: array of Variant); overload;
var
  n: integer;
begin
  for n:= Low(Params) to High(Params) do WriteOem(Params[n]);
end;

procedure WriteLnOem(const Value: Variant); overload;
begin
  WriteOem(Value);
  WriteLn;
end;

procedure WriteLnOem(const Params: array of Variant); overload;
begin
  WriteOem(Params);
  WriteLn;
end;

function WStrLen(PW: PWideChar): integer;
begin
  Result:= 0;
  if PW = nil then Exit;
  while PW^ <> #0 do
  begin
    Inc(PW);
    Inc(Result);
  end;
end;

function LoadStrRes(ResID: integer): string;
var
  Rec: TResStringRec;
begin
  Rec.Module:= @HINSTANCE;
  Rec.Identifier:= ResID;
  Result:= LoadResString(@Rec);
end;

function Redirected: boolean;
var
  StdOut: THandle;
  Mode: DWORD;
begin
  Result:= true;
  StdOut:= GetStdHandle(STD_OUTPUT_HANDLE);
  if INVALID_HANDLE_VALUE <> StdOut then
    Result:= not GetConsoleMode(StdOut, Mode);
end;

end.
